<?php
// Autor (c) Miroslav Novak, www.platiti.cz
// Pouzivani bez souhlasu autora neni povoleno
// #Ver:PRV101-56-g411a2eed:2023-07-24#

namespace Platiti\PtiUniNs\UniModul;

use Throwable;
use const Platiti\PtiUniNs\UniModul\ConstUniModulVarPath;

if(file_exists(ConstUniModulVarPath . '/UniModulConfig.php')) {
    include ConstUniModulVarPath . '/UniModulConfig.php';
}


// UniErr Logging
if (!defined('E_UNIERR_DEFAULT')) {
	define('E_UNIERR_DEFAULT', E_ALL & ~E_STRICT);
}

class UniErrBasic {

	/** @var UniLogger */
	protected $logger;
	/** @var bool */
	protected $diagLogTrace;

	/** @var ?string */
	public $specialShutdownLogMessage;

	/** @var array<int, int> */
	protected $levelsStack = [];
	/** @var bool */
	protected $requestHeaderPrinted;
	/** @var ?callable */
	protected $prevErrHandler;
	/** @var ?callable */
	protected $prevExcHandler;

	const E_UNIERR_INTERNAL = 2<<18;


	/**
	 * @param UniLogger $logger
	 * @param bool $enable
	 * @param bool $diagLogTrace - pro ladeni neparovych Begin/End UniErru
	 * @return void
	 */
	public function __construct($logger, $enable, $diagLogTrace) {
		$this->logger = $logger;
		$this->diagLogTrace = $diagLogTrace;
		if ($enable) {
			$this->prevErrHandler = set_error_handler([$this, "UniErrHandler"]);
			$this->prevExcHandler = set_exception_handler ([$this, "UniErrExceptionHandler"]);
			register_shutdown_function([$this, "UniErrShutdownHandler"]);
		}
	}

	/**
	 * @param int $erlev
	 * @return void
	 */
	public function begin($erlev = E_UNIERR_DEFAULT) {

		$this->levelsStack[] = $erlev;
		if ($this->diagLogTrace) $this->writeErrLog(static::E_UNIERR_INTERNAL, "BeginUniErr ".count($this->levelsStack), "0", 0, 2);
	}

	/**
	 * @template _T
	 * @param _T $ret
	 * @return _T
	 */
	public function end($ret = NULL) {
		if ($this->diagLogTrace) $this->writeErrLog(static::E_UNIERR_INTERNAL, "EndUniErr ".count($this->levelsStack), "0", 0, 2);
		if (count($this->levelsStack)!=0) {
			array_pop($this->levelsStack);
		} else {
			$this->writeErrLog(static::E_UNIERR_INTERNAL, "EndUniErr bez odpovidajiciho BeginUniErr", "0", 0 , 2);
			user_error("EndUniErr bez odpovidajiciho BeginUniErr");
		}
		return $ret;
	}

	/**
	 * @param array<int, mixed> $new
	 * @return array<int, mixed>
	 */
	public function reset($new = []) {
		if ($this->diagLogTrace) $this->writeErrLog(static::E_UNIERR_INTERNAL, "ResetUniErr ".count($this->levelsStack), "0", 0, 2);
		if (count($this->levelsStack)!=0) {
			$old = $this->levelsStack;
		} else {
			$old = [];
		}
		$this->levelsStack = $new;
		return $old;
	}

	/**
	 * @param Throwable $ex
	 * @return never
	 */
	public function exception($ex) {
		if (count($this->levelsStack)!=0) {
			$this->writeErrLog(0, 'Rethrowing exception '. $ex->__toString() . "\n" . $ex->getTraceAsString(), "0",  0);
		}
		throw $ex;
	}


	/**
	 * @param int $errno
	 * @param string $errstr
	 * @param ?string $errfile
	 * @param ?int $errline
	 * @param int $traces_to_ignore
	 * @param int|null $backTrace
	 * @return void
	 */
	public function writeErrLog($errno, $errstr, $errfile, $errline , $traces_to_ignore=3, $backTrace=null) {
		if (!$this->requestHeaderPrinted) {
			$this->logger->WriteLog("UniErrInfo: "."-------------------------------------------- new request #Ver:PRV101-56-g411a2eed:2023-07-24#");
			if (PHP_SAPI !== 'cli') {
				$this->logger->writeLogNoNewLines("UniErrInfo: "."REQUEST ". Http::server('REMOTE_ADDR')." ".Http::server('REQUEST_URI')." PostData: ".var_export($_POST, true));
			} else {
				$this->logger->writeLogNoNewLines("UniErrInfo: "."CLI Cmd: ".var_export($_SERVER['argv'], true));
			}
			$this->requestHeaderPrinted = true;
		}
		$this->logger->WriteLog("UniErrInfo: ".$this->errnoName($errno).", $errstr, $errfile, $errline\n" . ($backTrace!=null ? $backTrace : $this->get_debug_print_backtrace($traces_to_ignore)));
	}



	/**
	 * @param int $errno
	 * @param string $errstr
	 * @param string $errfile
	 * @param int $errline
	 * @return bool
	 */
	public function UniErrHandler($errno, $errstr, $errfile, $errline) {
		if (count($this->levelsStack)!=0) {
			$erlev = end($this->levelsStack);
			if ($erlev & $errno) {
				$this->writeErrLog($errno, $errstr, $errfile, $errline );
			}
		}
		if ($this->prevErrHandler !== null) {
			return (bool)call_user_func($this->prevErrHandler, $errno, $errstr, $errfile, $errline); /** #phpstan-ignore-line */
		} else {
			return false; // normal err handling
		}
	}

	/**
	 * @param Throwable $ex
	 * @return void
	 */
	public function UniErrExceptionHandler($ex) {
		if (count($this->levelsStack)!=0) {
			$this->writeErrLog(0, 'Unhandled exception ' .  $ex->__toString(), "0",  0);
		}
		if ($this->prevExcHandler !== null) {
			call_user_func($this->prevExcHandler, $ex);
		} else {
			$ue = $this->reset();
			user_error($ex->__toString());
			$this->reset($ue);
		}
	}

	/** @return void  */
	public function UniErrShutdownHandler() {
		if (count($this->levelsStack)!=0) {
			$err = error_get_last();
			if ($err != null && $err['type'] & (E_ERROR | E_PARSE | E_CORE_ERROR | E_COMPILE_ERROR | E_USER_ERROR | E_RECOVERABLE_ERROR)) {
				$this->prevErrHandler = null; //vypneme volani pripadneho retezeneho err handleru
				$this->UniErrHandler($err['type'], 'UniErr-SHUTDOWN ' . $err['message'], $err['file'],  $err['line']);
			} else {
				if ($this->specialShutdownLogMessage !== null) {
					$this->writeErrLog(static::E_UNIERR_INTERNAL, "SpecialShutdown: " . $this->specialShutdownLogMessage, "0", 0);
				} else {
					$this->writeErrLog(static::E_UNIERR_INTERNAL, "Shutdown s aktivním UniErrem!, hloubka: ".count($this->levelsStack), "0", 0);
				}
			}
			$this->reset(); // musi se to killnout, jinak nasledny shutdown kod vyvolava ruzne warningu, coz je matouci
		}


	}

	/**
	 * @param int $traces_to_ignore
	 * @return string
	 */
	protected function get_debug_print_backtrace($traces_to_ignore = 1) {
		$traces = debug_backtrace();
		$ret = array();
		foreach($traces as $i => $call){
			if ($i < $traces_to_ignore ) {
				continue;
			}

			$object = '';
			$x = $call['function'];
			if (isset($call['class'])) {
				$object = $call['class'] . (isset($call['type']) ? $call['type'] : '');
				/*
				if (isset($call['args']) && is_array($call['args'])) {
					foreach ($call['args'] as &$arg) {
						get_arg($arg);
					}
				}
				*/
			}
			if (isset($call['file']) && isset($call['line']) && isset($call['args'])) {
				$ret[] = '#'.str_pad(strval($i - $traces_to_ignore), 3, ' ').$object.$call['function'].'('./*implode(', ', $call['args']).*/') called at ['.$call['file'].':'.$call['line'].']';
			}
		}

		return implode("\n",$ret);
	}

	/*  // TODO: sice se nepouziva, ale kdyby jo, nutno odstranit dosazovani do reference $arg!!
	function get_arg(&$arg) {
		if (is_object($arg)) {
			$arr = (array)$arg;
			$args = array();
			foreach($arr as $key => $value) {
				if (strpos($key, chr(0)) !== false) {
					$key = '';    // Private variable found
				}
				$args[] =  '['.$key.'] => '.get_arg($value);
			}

			$arg = get_class($arg) . ' Object ('.implode(',', $args).')';
			STOPPP('funkce prave zmenila promenou v ramci stacku, ma dopad na runtime reference parametry!');
		}
	}
	*/

	/**
	 * @param int $errno
	 * @return string
	 */
	protected function errnoName($errno) {
		$errbitNames =  array('E_ERROR', 'E_WARNING', 'E_PARSE', 'E_NOTICE', 'E_CORE_ERROR', 'E_CORE_WARNING', 'E_COMPILE_ERROR', 'E_COMPILE_WARNING', 'E_USER_ERROR', 'E_USER_WARNING', 'E_USER_NOTICE', 'E_STRICT', 'E_RECOVERABLE_ERROR', 'E_DEPRECATED', 'E_USER_DEPRECATED', 'E_UNIERR_INTERNAL');
		$errStr = "";
		for ($i=0; $i<count($errbitNames); $i++) {
			if (($errno >> $i) & 1) {
				$errStr .= $errbitNames[$i] . " ";
			}
		}
		return $errStr;
	}
}



class UniErr {
	/** @var UniErrBasic */
	protected static $uniErr;


	/**
	 * @param int $erlev
	 * @return void
	 */
	public static function begin($erlev = E_UNIERR_DEFAULT) {
		static::getUniErr()->begin();
	}

	/**
	 * @template _T
	 * @param _T $ret
	 * @return _T
	 */
	public static function end($ret = NULL) {
		return static::getUniErr()->end($ret);
	}

	/**
	 * @param array<int, mixed> $new
	 * @return array<int, mixed>
	 */
	public static function reset($new = []) {
		return static::getUniErr()->reset($new);
	}

	/**
	 * @param Throwable $ex
	 * @return never
	 */
	public static function exception($ex) {
		static::getUniErr()->exception($ex);
	}

	/**
	 * @param int $errno
	 * @param string $errstr
	 * @param ?string $errfile
	 * @param ?int $errline
	 * @param int $traces_to_ignore
	 * @param int|null $backTrace
	 * @return void
	 */
	public static function writeErrLog($errno, $errstr, $errfile, $errline , $traces_to_ignore=3, $backTrace=null) {
		static::getUniErr()->writeErrLog($errno, $errstr, $errfile, $errline , $traces_to_ignore, $backTrace);
	}



	/** @return UniErrBasic */
	protected static function getUniErr() {
		if (static::$uniErr === null) {
			static::$uniErr = new UniErrBasic(new UniLogger(), !empty($GLOBALS['UniErrorControl']), !empty($GLOBALS['UniErrorReporting_DBG']));
		}
		return static::$uniErr;
	}
}


/**
 * @param int $erlev
 * @return void
 */
function BeginUniErr($erlev = E_UNIERR_DEFAULT) {
	UniErr::begin($erlev);
}

/**
 * @template _T
 * @param _T $ret
 * @return _T
 */
function EndUniErr($ret = NULL) {
	return UniErr::end($ret);
}

/**
 * @param array<int, mixed> $new
 * @return array<int, mixed>
 */
function ResetUniErr($new = array()) {
	return UniErr::reset($new);
}

/**
 * @param Throwable $ex
 * @return never
 */
function ExceptionUniErr($ex) {
	UniErr::exception($ex);
}

/**
 * @param int $errno
 * @param string $errstr
 * @param ?string $errfile
 * @param ?int $errline
 * @param int $traces_to_ignore
 * @param int|null $backTrace
 * @return void
 */
function UniWriteErrLog($errno, $errstr, $errfile, $errline , $traces_to_ignore=3, $backTrace=null) {
	UniErr::writeErrLog($errno, $errstr, $errfile, $errline , $traces_to_ignore, $backTrace);
}




/**
 * @param mixed $v
 * @param string $n
 * @return void
 */
function my_dump($v, $n = '') {
	echo "<p>$n <pre>";
	var_export($v);
	echo "</pre></p>";
}


